static char rcsid[] = "@(#)$Id: headers.c,v 1.4 2000/12/03 15:22:37 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.4 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "hdr_imp.h"
#include "s_me.h"

static struct header_info header_types[] = {
    /* From STD 11 (RFC 822): */
    { "Subject",    HDR_TEXT },
    { "Comments",   HDR_TEXT },
    { "Return-path",HDR_STRUCTURED },
    { "Received",   HDR_STRUCTURED },
    { "Reply-To",   HDR_STRUCTURED|HDR_PHRASE },    /* Not exactly correct .. */
    { "From",       HDR_STRUCTURED|HDR_PHRASE },    /* Not exactly correct .. */
    { "Sender",     HDR_STRUCTURED|HDR_PHRASE },    /* Not exactly correct .. */
    { "Resent-Reply-To", HDR_STRUCTURED|HDR_PHRASE },/* Not exactly correct .. */
    { "Resent-From",     HDR_STRUCTURED|HDR_PHRASE },/* Not exactly correct .. */
    { "Resent-Sender",   HDR_STRUCTURED|HDR_PHRASE },/* Not exactly correct .. */
    { "Date",        HDR_STRUCTURED },
    { "Resent-Date", HDR_STRUCTURED },
    { "To",          HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "Resent-To",   HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "cc",          HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "Resent-cc",   HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "bcc",         HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "Resent-bcc",  HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "Message-ID",  HDR_STRUCTURED },
    { "Resent-Message-ID",  HDR_STRUCTURED },
    { "In-Reply-To", HDR_STRUCTURED|HDR_PHRASE }, /* Is correct .. */
    { "References",  HDR_STRUCTURED|HDR_PHRASE }, /* Is correct .. */
    { "Keywords",  HDR_STRUCTURED|HDR_PHRASE },   /* Is correct :-) .. */
    { "Encrypted", HDR_STRUCTURED },              /* Well ... */
    /* From MIME (RFC 1521) */
    { "MIME-Version",              HDR_STRUCTURED },     
    { "Content-Type",              HDR_STRUCTURED },     
    { "Content-Transfer-Encoding", HDR_STRUCTURED },
    { "Content-ID",                HDR_STRUCTURED },     
    { "Content-Description",       HDR_TEXT },
    /* From RFC 1806 */
    { "Content-Disposition",       HDR_STRUCTURED },
    /* From RFC 1864 */
    { "Content-MD5",               0 },
    
    /* mailbox format */
    { "Content-Length",      0 },
    { "Status",              0 },

    /* Sendmail */
    { "Full-Name",           HDR_TEXT },
    { "Return-Receipt-To",   HDR_STRUCTURED|HDR_PHRASE }, /* Not exactly correct .. */
    { "Auto-Submitted",      HDR_STRUCTURED },
    { "Precedence",          0 },
    

    /* IDA Sendmail */
    { "X-Charset",           0 },
    { "X-Char-Esc",          0 },
    
    /* Unknown source */
    { "Action",              0 },
    { "Priority",            HDR_STRUCTURED },
    { "Expires",             HDR_STRUCTURED },
    { "Importance",          HDR_STRUCTURED },
    { "Sensitivity",         HDR_STRUCTURED },

    /* Our non-standard headers */
    { "X-ELM-OSV",                 HDR_STRUCTURED },
    { "X-Mailer",                  HDR_TEXT },
    { "Content-Name",              HDR_TEXT },
  
    /* Tailer */
    { NULL, 0 }
};

/* Grows forever */
static struct hdr_list {
    struct hdr_list * next;
    struct header_info hdr;
} * extra_headers = NULL;
static int extra_header_count = 0;

header_ptr find_header(name, create_flag) 
     CONST char *name;
     int create_flag;
{
    int i;
    struct hdr_list * walk, *last = NULL;
    
    for (i = 0; header_types[i].header; i++) 
	if (strincmp(name,header_types[i].header,STRING) == 0)
	    return &(header_types[i]);
    
    for (walk = extra_headers; 
	 walk != NULL; 
	 last = walk, walk = walk -> next)
	if (strincmp(name,walk->hdr.header,STRING) == 0)
	    return &(walk->hdr);
    
    if (create_flag) {
	walk = (struct hdr_list * ) safe_malloc (sizeof(struct hdr_list));
	
	if (last) last->next = walk;
	else extra_headers   = walk;
	extra_header_count++;
	
	walk->next = NULL;
	walk->hdr.header = strmcpy(NULL,name);
	
	if (strincmp(name,"X-",2) == 0)
	    walk->hdr.flag = HDR_TEXT;     /* Default for user defined headers */
	else
	    walk->hdr.flag = 0;            /* We don't know it, right? */
	
	return &(walk->hdr);
    }
  
    return NULL;
}

int classify_header(name)
     CONST char *name;
{
    header_ptr P = find_header(name,0);

    if (P)
	return P->flag;
    
    if (strincmp(name,"X-",2) == 0)
	return HDR_TEXT;              /* Default for user defined headers */

    return 0;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


header_list_ptr locate_header(h,n)
     header_list_ptr h;
     header_ptr n;
{
    header_list_ptr walk;

    for (walk = h; 
	 walk != NULL; 
	 walk = walk -> next_other_header) {
	if (walk -> magic != HEADER_magic)
	    panic("HEADERS PANIC",__FILE__,__LINE__,"locate_header",
		  "Bad magic number",0);
	if (n == walk -> header_name)
	    return walk;
    }
    return NULL;
}

void update_header_list(list,last1, name, value)
     header_list_ptr *list;
     header_list_ptr *last1;
     CONST char * name;
     CONST char * value;
{
    header_list_ptr result = *list, last = *last1;
    header_list_ptr item, last_this = NULL, walk;

    if (result && !last) {
	last = result;

	dprint(12,
	       (debugfile,
		"update_header_list: Moving 'last' to end of list ... \n"));

	while (last -> next_header)
	    last = last -> next_header;

    }

    item = (struct header_list *) safe_malloc(sizeof (struct header_list));
    item -> header_name = find_header(name,1);
	
    item -> next_header       = NULL;
    item -> next_this_header  = NULL;
    item -> next_other_header = NULL;
    item -> body              = safe_strdup(value);
    item -> magic             = HEADER_magic;
	
    for (walk = result; 
	 walk != NULL; 
	 last_this = walk, walk = walk -> next_other_header) {
	if (walk -> magic != HEADER_magic)
	    panic("HEADERS PANIC",__FILE__,__LINE__,"update_header_list",
		  "Bad magic number",0);
	if (item -> header_name == walk -> header_name)
	    break;
    }
	
    if (!walk) {
	dprint(12, (debugfile,
		    "update_header_list: name='%s': (%p) Not in chain!\n",
		    item -> header_name->header,
		    item -> header_name));
    }

    if (walk) {  
	while (walk->next_this_header != NULL) {
	    if (walk -> magic != HEADER_magic)
		panic("HEADERS PANIC",__FILE__,__LINE__,"update_header_list",
		      "Bad magic number",0);
	    walk = walk->next_this_header;
	}
	    
	walk->next_this_header = item;
	
	dprint(12, (debugfile,
		    "update_header_list: name='%s' -- append next_this_header (%s)\n",
		    item->header_name->header,
		    walk->header_name->header));
	    	
    } else if (last_this) {
	last_this -> next_other_header = item;
	dprint(12,(debugfile,
		   "update_header_list: name='%s' -- append next_other_header (%s)\n",
		   item->header_name->header,
		   last_this->header_name->header));
    }
    
    if (last) {
	last -> next_header  = item;
	
	dprint(12, (debugfile,
		    "update_header_list: name='%s' -- append next_header (%s)\n",
		    item->header_name->header,
		    last->header_name->header));
	
    } else  {
	result               = item;
	dprint(12, (debugfile,
		    "update_header_list: name='%s' -- head of next_header\n",
		    item->header_name->header));
    }
    last = item;
    
    *list  = result;
    *last1 = last;
}

void delete_headers(hdr)
     header_list_ptr *hdr;
{
    header_list_ptr next = *hdr;
  
    while(next) {
	if (next -> magic != HEADER_magic)
	    panic("HEADERS PANIC",__FILE__,__LINE__,"delete_headers",
		  "Bad magic number",0);
	
	(*hdr)  = next;
	next = next -> next_header;
	
	if((*hdr) -> body) {
	    free((*hdr) -> body);
	    (*hdr) -> body = NULL;
	}
	(*hdr) -> next_header       = NULL;
	(*hdr) -> next_this_header  = NULL;
	(*hdr) -> next_other_header = NULL;
	(*hdr) -> magic             = 0;
	free((void *)(*hdr));
    }
    (*hdr)  = next;
}

header_list_ptr read_folder_headers(read_state_ptr,folder,entry)
     READ_STATE read_state_ptr;
     struct folder_info *folder;
     struct header_rec *entry;
{
    char *buffer;
    int len;

    header_list_ptr result = NULL, last = NULL;
    header_list_ptr tmphdr;

    while (copy_header_folder(folder,read_state_ptr, &buffer,&len)) {
	char * val;
	
	unfold_header(buffer,&len,entry);
	val = strchr(buffer,':');

	if (!val) {
	    panic("HEADERS PANIC",__FILE__,__LINE__,"read_folder_headers",
		  "bad return from copy_header_folder -- ':' not found",0);
	}

	*val = '\0';
	val++;

	while (whitespace(*val))
	    val++;

	no_ret(val);

	update_header_list(&result,&last,buffer,val);

	free(buffer); buffer = NULL;
    }

    if (NULL != (tmphdr = locate_header_by_name(result,"Return-Path"))) {
	/* Required for POP or 
	 * when env_from_source == 2
	 */
	if (!entry->env_from[0]) {
	    return_path_to_env_from(entry,tmphdr->body);
	}
    }	

    if (NULL != (tmphdr = locate_header_by_name(result,"Message-Id"))) {
	strfcpy(entry->messageid, tmphdr->body, 
		sizeof entry->messageid);
    }

    if (NULL != (tmphdr = locate_header_by_name(result,"Content-Length"))) {
	entry->content_length = atol(tmphdr->body);
    }

    if (NULL != (tmphdr = locate_header_by_name(result,"Expires"))) {
	process_expiration_date(tmphdr->body, &(entry->status));
    }

    if (NULL != (tmphdr = locate_header_by_name(result,"Date"))) {
	parse_arpa_date(tmphdr->body, entry);
    }
    
    /* We should accept headers like:
     *    MIME-Version: (MetaSend v1.7) 1.0
     * And if version number is something other accept it but give 
     * warning when message is viewed or replied.
     */

    if (NULL != (tmphdr = locate_header_by_name(result,"MIME-Version"))) {

	if (0 == strcmp(tmphdr->body,"1.0"))
	    entry->status |= MIME_MESSAGE;
	else {
	    entry->status |= MIME_MESSAGE|MIME_UNSUPPORTED;
	    dprint(12, (debugfile, 
			"read_folder_headers: Unsupported mime version '%s'\n",
			tmphdr->body));
	}

	if (tmphdr->next_this_header &&
	    0 != istrcmp(tmphdr->body,tmphdr->next_this_header->body) &&
            show_header_errors) {
            lib_error(CATGETS(elm_msg_cat, MeSet, MeParseErrorMIMEVersion,
                              "PARSE ERROR: Several MIME-Version headers!"));
        }


    }    

    return result;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */

